#!/usr/bin/env ruby

require "rxldata.rb"
require "rxlgraph.rb"
require "rxleval.rb"

class Rxl < Object

HELP = <<text
rxl - ruby excel replacement on the command line

Commands:
h - show help
<A..Z><0..N> - cell (column, row)
s - show evaluated sheet
e - show entered sheet 
e <cell> <data> - enter cell data
<data>,<data>,...<enter> add data for row and edit in next row
<data>;<data>;...<enter> add data for col and edit in next col

Cell fomatting:
"string"
1234, 0x1220 - numbers
<a1> - cell reference
<a1>-<a3> - row/column range
<a1>,<a2>, ... - cell set

Cell expressions:
+,-,*,/,** - evaluated in ruby priority order
<expr> ? <true_expr> : <false_expr>
sum(<cells>), mean(<cells>), var(<cells>), min(<cells>), max(<cells>)
Math::E, Math::PI - constants
1.0, 1234, 0x1220 - numbers
"text"
any other Ruby expression

Future:
- Add ReadLine support (http://bogojoker.com/readline/)

Example:
$ rxl.sh
$ rxl>> e a1 1,2
$ rxl>> e a2 2,2
$ rxl>> s
$ rxl>> e a3 sum(<a1>,<a2>),sum(<b1>,<b2>)
$ rxl>> s
$ rxl>> e c1 sum(<a1>,<b1>);sum(<a2>,<b2>);<c1>+<c2>
$ rxl>> s
$ rxl>> q

text
  
  CMD = { :help => /^h/, :show => /^s/, :edit => /^e/, :quit => /^q/, :input => /^i/, :output => /^o/ }
  COMMA_SEP = /([^,\(]*(\([^\)]*\)[^,\(]*)*)/	  # do not split within parentheses
  SPACE_SEP = /"([^"]+)"|(\S+)/                 # handle "1 2" 3

  def initialize()
	  @data = RxlData.new
	  @cell = ""
  end
 
  def parse(s)
    cmd, param, *data = s.scan(SPACE_SEP).flatten.compact
    data = data.join().gsub(/\s/,'')
  	case cmd
  	when CMD[:help]: puts HELP
  	when CMD[:show]: print_eval
    when CMD[:edit]: (param and data) ? edit(param, data) : print_data
  	when CMD[:quit]: quit
  	when CMD[:input]: input(param)
  	when CMD[:output]: print_data(param)
  	else puts "Dont understand command"
  	end
  end

  # Data may be a single value or a comma/semicolon separated data
  # val row: <data>, <data>, ...
  # val column: <data>; <data>; ...
  def edit(cells, data)
    c, r = RxlData::parse(cells)
    if data =~ /[^;]+(;[^;]+)+/
      data.scan(/[^;]+/).each { |v| @data.set(c+r, v.strip); r = r.next }
    elsif data =~ COMMA_SEP
      Rxl.comma_split(data).each { |v| @data.set(c+r, v.strip); c = c.next }
    else
      @data.set(cells, data.strip)
    end
  end

  # Print entered data
  def print_data(data=nil, file=nil)
    data = data || @data
    o = file ? File.new(file, "w") : IO.new(1,"w")
    data.rows.each { |r|
      rstr = ""
      data.cols.each { |c|
        i = data.get(c+r)
	      rstr = rstr + (i ? i.to_s : "-") + ", "
	    }
	    o.puts rstr.chop
	  }
  end
  
  # Print evaluated data
  def print_eval(file=nil)
    o = file ? File.new(file, "w") : IO.new(1,"w")
	  graph = RxlGraph.new
	  cells = []
	  tmp = @data.clone			# Deep clone
	  # convert ranges/functions and build dependency graph 
	  @data.each { |c,v|
	    cells << c
	    nv = RxlEval.convert_ranges(v)
	    nv = RxlEval.convert_functions(nv)
        RxlEval.cells(nv).each { |to| graph.add(c,to) }
	    tmp.set(c, nv)
	  }
    deps = nil
    until graph.empty? || (graph.dependencies == deps)
      deps = graph.dependencies
  	  graph.no_out_dep(cells).each { |to|
  	    graph.dep_to(to).each { |from|
  	      tmp.set(from, RxlEval.expand_params(tmp.get(from), { to => tmp.get(to) }))
  	      graph.remove(from,to)
  	    }
  	  }
	  end
    if graph.empty?
      tmp.each { |c,v| tmp.set(c, eval(v)) }
      print_data(tmp)
    else
      puts "Failed to evaluate sheet. Check for cyclic dependencies"
    end
  end

  def input(file=nil)
    # TBD
  end
  
  def quit
    # TODO: Add confirmation
    exit
  end
  
  def Rxl.comma_split(s)
    ans = s.scan(COMMA_SEP).collect { |n| n[0] }
    ans = ans.reject { |n| (!n || n.size == 0) }
  end
  
  def Rxl.one_of(x, ary)
  end
  
  # Main
  rxl = Rxl.new
  while true 
    print "rxl>> "
    s = gets
    rxl.parse(s)
  end

end
